/*
 * Sweep, a sound wave editor.
 *
 * Copyright (C) 2000 Conrad Parker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <pthread.h>

#include <config.h>

#include <sndfile.h>

#define BUFFER_LEN 1024

#include <glib.h>
#include <gtk/gtk.h>

#include <sweep/sweep_i18n.h>
#include <sweep/sweep_types.h>
#include <sweep/sweep_typeconvert.h>
#include <sweep/sweep_sample.h>
#include <sweep/sweep_undo.h>

#include "sample.h"
#include "file_dialogs.h"
#include "file_sndfile.h"
#include "view.h"

#define BUF_LEN 128

#if ! defined (SNDFILE_1)

typedef struct {
  int subformat;
  char * name;
} sndfile_subformat_names;

static sndfile_subformat_names subformat_names[] = {
  { SF_FORMAT_PCM, "PCM (Pulse Code Modulation)", },
  { SF_FORMAT_FLOAT, "32 bit floating point", },
  { SF_FORMAT_ULAW, "U-Law", },
  { SF_FORMAT_ALAW, "A-Law", },
  { SF_FORMAT_IMA_ADPCM, "IMA ADPCM", },
  { SF_FORMAT_MS_ADPCM, "Microsoft ADPCM", },
  { SF_FORMAT_PCM_BE, "Big endian PCM", },
  { SF_FORMAT_PCM_LE, "Little endian PCM", },
  { SF_FORMAT_PCM_S8, "Signed 8 bit PCM", },
  { SF_FORMAT_PCM_U8, "Unsigned 8 bit PCM", },
  { SF_FORMAT_SVX_FIB, "SVX Fibonacci Delta", },
  { SF_FORMAT_SVX_EXP, "SVX Exponential Delta", },
  { SF_FORMAT_GSM610, "GSM 6.10", },
  { SF_FORMAT_G721_32, "32kbs G721 ADPCM", },
  { SF_FORMAT_G723_24, "24kbs G723 ADPCM", },
  { SF_FORMAT_FLOAT_BE, "Big endian floating point", },
  { SF_FORMAT_FLOAT_LE, "Little endian floating point", },
};

/*
 * Subformat capabilities.
 *
 * Logic for these derived from sf_format_check() in libsndfile::sndfile.c
 */

typedef struct {
  int subformat;
  gboolean cap_8bit;
  gboolean cap_16bit;
  gboolean cap_24bit;
  gboolean cap_32bit;
  gboolean cap_mono;
  gboolean cap_stereo;
} sndfile_subformat_caps;

static sndfile_subformat_caps wav_caps[] = {
  { SF_FORMAT_PCM, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_IMA_ADPCM, FALSE, TRUE, FALSE, FALSE, TRUE, TRUE, },
  { SF_FORMAT_MS_ADPCM, FALSE, TRUE, FALSE, FALSE, TRUE, TRUE, },
  { SF_FORMAT_GSM610, FALSE, TRUE, FALSE, FALSE, TRUE, FALSE, },
  { SF_FORMAT_ULAW, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_ALAW, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_FLOAT, FALSE, FALSE, FALSE, TRUE, TRUE, TRUE, },
  { 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE},
};

static sndfile_subformat_caps aiff_caps[] = {
  { SF_FORMAT_PCM, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_PCM_LE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_PCM_BE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_FLOAT, FALSE, FALSE, FALSE, TRUE, TRUE, TRUE, },
  { 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE},
};

static sndfile_subformat_caps au_caps[] = {
  { SF_FORMAT_PCM, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_ULAW, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_ALAW, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_G721_32, TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, },
  { SF_FORMAT_G723_24, TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, },
  { SF_FORMAT_FLOAT, FALSE, FALSE, FALSE, TRUE, TRUE, TRUE, },
  { 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE},
};

static sndfile_subformat_caps raw_caps[] = {
  { SF_FORMAT_PCM_S8, TRUE, FALSE, FALSE, FALSE, TRUE, TRUE, },
  { SF_FORMAT_PCM_U8, TRUE, FALSE, FALSE, FALSE, TRUE, TRUE, },
  { SF_FORMAT_PCM_BE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_PCM_LE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE},
};

static sndfile_subformat_caps paf_caps[] = {
  { SF_FORMAT_PCM_S8, TRUE, FALSE, FALSE, FALSE, TRUE, TRUE, },
  { SF_FORMAT_PCM_BE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_PCM_LE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE},
};

static sndfile_subformat_caps svx_caps[] = {
  { SF_FORMAT_PCM, TRUE, TRUE, FALSE, FALSE, TRUE, TRUE },
  { 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE},
};

static sndfile_subformat_caps nist_caps[] = {
  { SF_FORMAT_PCM, FALSE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_PCM_BE, FALSE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_PCM_LE, FALSE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE},
};

static sndfile_subformat_caps ircam_caps[] = {
  { SF_FORMAT_PCM, FALSE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_PCM_BE, FALSE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_PCM_LE, FALSE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_FLOAT, FALSE, FALSE, FALSE, TRUE, TRUE, TRUE, },  
  { SF_FORMAT_FLOAT_BE, FALSE, FALSE, FALSE, TRUE, TRUE, TRUE, },  
  { SF_FORMAT_FLOAT_LE, FALSE, FALSE, FALSE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_ULAW, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { SF_FORMAT_ALAW, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, },
  { 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE},
};

typedef struct {
  int format;
  sndfile_subformat_caps * subformat_caps;
} sndfile_format_caps;

static sndfile_format_caps format_caps[] = {
  { SF_FORMAT_WAV, wav_caps, },
  { SF_FORMAT_AIFF, aiff_caps, },
  { SF_FORMAT_AU, au_caps, },
  { SF_FORMAT_AULE, au_caps, },
  { SF_FORMAT_RAW, raw_caps },
  { SF_FORMAT_PAF, paf_caps, },
  { SF_FORMAT_SVX, svx_caps, },
  { SF_FORMAT_NIST, nist_caps, },
  { SF_FORMAT_IRCAM, ircam_caps, },
};

typedef struct {
  sw_sample * sample; /* used for saving */
  char * pathname; /* used for loading */
  SF_INFO * sfinfo;
  GtkWidget * radio_mono;
  GtkWidget * radio_stereo;
  GtkWidget * radio_8bit;
  GtkWidget * radio_16bit;
  GtkWidget * radio_24bit;
  GtkWidget * radio_32bit;
  GtkWidget * ok_button;
} sndfile_save_options;

static char *
get_subformat_name (int subformat)
{
  int i;

  subformat &= SF_FORMAT_SUBMASK;

  for (i = 0; i < sizeof(subformat_names)/sizeof(sndfile_subformat_names); i++) {
    if (subformat_names[i].subformat == subformat)
      return subformat_names[i].name;
  }

  return NULL;
}

static sndfile_subformat_caps *
get_format_caps (int format)
{
  int i;

  format &= SF_FORMAT_TYPEMASK;

  for (i = 0; i < sizeof(format_caps)/sizeof(sndfile_format_caps); i++) {
    if (format_caps[i].format == format)
      return format_caps[i].subformat_caps;
  }

  return NULL;
}

static sndfile_subformat_caps *
get_subformat_caps (sndfile_subformat_caps * format_subcaps, int subformat)
{
  subformat &= SF_FORMAT_SUBMASK;

  while (format_subcaps) {
    if (format_subcaps->subformat == subformat) {
      return format_subcaps;
    }
    format_subcaps++;
  }

  return NULL;
}

static void
sndfile_save_options_dialog_ok_cb (GtkWidget * widget, gpointer data)
{
  sndfile_save_options * so = (sndfile_save_options *)data;
  sw_sample * sample = so->sample;
  GtkWidget * dialog;

  dialog = gtk_widget_get_toplevel (widget);
  gtk_widget_destroy (dialog);

  if (sample->file_info) {
    g_free (sample->file_info);
  }

  sample->file_info = so->sfinfo;

  g_free (so);

  sndfile_sample_save (sample);
}

static void
sndfile_save_options_dialog_cancel_cb (GtkWidget * widget, gpointer data)
{
  sndfile_save_options * so = (sndfile_save_options *)data;
  GtkWidget * dialog;

  dialog = gtk_widget_get_toplevel (widget);
  gtk_widget_destroy (dialog);

  g_free (so);
}

static void
update_ok_button (sndfile_save_options * so)
{
  SF_INFO * sfinfo = so->sfinfo;

  gtk_widget_set_sensitive (so->ok_button, sf_format_check (sfinfo));
}

static void
update_save_options_caps (sndfile_save_options * so)
{
  SF_INFO * sfinfo;
  sndfile_subformat_caps * fcaps, * caps;

  sfinfo = so->sfinfo;
  fcaps = get_format_caps (sfinfo->format);
  caps = get_subformat_caps (fcaps, sfinfo->format);

  if (caps) {
    gtk_widget_set_sensitive (so->radio_mono, caps->cap_mono);
    gtk_widget_set_sensitive (so->radio_stereo, caps->cap_stereo);
    gtk_widget_set_sensitive (so->radio_8bit, caps->cap_8bit);
    gtk_widget_set_sensitive (so->radio_16bit, caps->cap_16bit);
    gtk_widget_set_sensitive (so->radio_24bit, caps->cap_24bit);
    gtk_widget_set_sensitive (so->radio_32bit, caps->cap_32bit);
  } else {
    gtk_widget_set_sensitive (so->radio_mono, FALSE);
    gtk_widget_set_sensitive (so->radio_stereo, FALSE);
    gtk_widget_set_sensitive (so->radio_8bit, FALSE);
    gtk_widget_set_sensitive (so->radio_16bit, FALSE);
    gtk_widget_set_sensitive (so->radio_24bit, FALSE);
    gtk_widget_set_sensitive (so->radio_32bit, FALSE);
  }

  update_ok_button (so);
}

static void
update_save_options_values (sndfile_save_options * so)
{
  sw_sample * sample;
  sw_format * format;
  SF_INFO * sfinfo;

  sample = so->sample;
  format = sample->sounddata->format;

  sfinfo = so->sfinfo;

#define BLOCK_AND_SET(r,d) \
  gtk_signal_handler_block_by_data(GTK_OBJECT((r)), (d));       \
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON((r)), TRUE);  \
  gtk_signal_handler_unblock_by_data(GTK_OBJECT((r)), (d));

  switch (sfinfo->pcmbitwidth) {
  case 8:
    BLOCK_AND_SET(so->radio_8bit, so);
    break;
  case 16:
    BLOCK_AND_SET(so->radio_16bit, so);
    break;
  case 24:
    BLOCK_AND_SET(so->radio_24bit, so);
    break;
    case 32:
      BLOCK_AND_SET(so->radio_32bit, so);
      break;
  default:
    break;
  }

  switch (sfinfo->channels) {
  case 1:
    BLOCK_AND_SET(so->radio_mono, so);
    break;
  case 2:
    BLOCK_AND_SET(so->radio_stereo, so);
    break;
  default:
    break;
  }

  update_ok_button (so);
}

static void
set_subformat_cb (GtkWidget * widget, gpointer data)
{
  sndfile_save_options * so = (sndfile_save_options *)data;
  SF_INFO * sfinfo = so->sfinfo;
  int subformat;

  sfinfo->format &= SF_FORMAT_TYPEMASK; /* clear submask */

  subformat = (int) gtk_object_get_user_data (GTK_OBJECT(widget));
  subformat &= SF_FORMAT_SUBMASK; /* get new subformat */

  sfinfo->format |= subformat;

  update_save_options_caps (so);
}

static void
set_bits_cb (GtkWidget * widget, gpointer data)
{
  sndfile_save_options * so = (sndfile_save_options *)data;
  SF_INFO * sfinfo = so->sfinfo;

  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget))) {
    sfinfo->pcmbitwidth = (int) gtk_object_get_user_data (GTK_OBJECT(widget));
  }

  update_ok_button (so);

  return;
}

static void
set_channels_cb (GtkWidget * widget, gpointer data)
{
  sndfile_save_options * so = (sndfile_save_options *)data;
  SF_INFO * sfinfo = so->sfinfo;

  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget))) {
    sfinfo->channels = (int) gtk_object_get_user_data (GTK_OBJECT(widget));
  }

  update_ok_button (so);

  return;
}

static GtkWidget *
create_sndfile_save_options_dialog (sndfile_save_options * so)
{
  GtkWidget * dialog;
  GtkWidget * table;
  GtkWidget * hbox;
  GtkWidget * label;
  GtkWidget * option_menu;
  GtkWidget * menu;
  GtkWidget * menuitem;
  GtkWidget * ok_button, * button;
  GtkWidget * radio;

  sw_sample * sample = so->sample;
  SF_INFO * sfinfo = so->sfinfo;
  sndfile_subformat_caps * fcaps;
  int subformat;
  char * name;

  dialog = gtk_dialog_new ();
  gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
  gtk_container_border_width (GTK_CONTAINER(dialog), 8);

  label = gtk_label_new (so->pathname);
  gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->vbox), label,
		      FALSE, FALSE, 0);
  gtk_widget_show (label);

  table = gtk_table_new (4, 2, FALSE);
  gtk_table_set_row_spacings (GTK_TABLE(table), 8);
  gtk_table_set_col_spacings (GTK_TABLE(table), 8);
  gtk_box_pack_start (GTK_BOX(GTK_DIALOG(dialog)->vbox), table,
		      FALSE, FALSE, 0);
  gtk_widget_show (table);

  /* Encoding */

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_table_attach (GTK_TABLE(table), hbox, 0, 1, 0, 1,
		    GTK_FILL|GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show (hbox);
  
  label = gtk_label_new ("Encoding:");
  gtk_box_pack_end (GTK_BOX(hbox), label, FALSE, FALSE, 0);
  gtk_widget_show (label);

  option_menu = gtk_option_menu_new ();
  gtk_table_attach (GTK_TABLE(table), option_menu, 1, 2, 0, 1,
		    GTK_FILL|GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show (option_menu);

  menu = gtk_menu_new ();

  fcaps = get_format_caps (sfinfo->format);
  
  subformat = sfinfo->format & SF_FORMAT_SUBMASK;

  if (fcaps) {
    while (fcaps && (fcaps->subformat != 0)) {
      name = get_subformat_name(fcaps->subformat);
      menuitem = gtk_menu_item_new_with_label (name);
      gtk_object_set_user_data (GTK_OBJECT(menuitem),
				(gpointer)fcaps->subformat);
      gtk_signal_connect (GTK_OBJECT(menuitem), "activate",
			  GTK_SIGNAL_FUNC(set_subformat_cb), so);
      gtk_menu_append (GTK_MENU(menu), menuitem);
      gtk_widget_show (menuitem);
      if (fcaps->subformat == subformat) {
	gtk_menu_item_select (GTK_MENU_ITEM(menuitem));
      }
      fcaps++;
    }
  }

  gtk_option_menu_set_menu (GTK_OPTION_MENU(option_menu), menu);
  gtk_widget_show (menu);

  /* Rate */

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_table_attach (GTK_TABLE(table), hbox, 0, 1, 1, 2,
		    GTK_FILL|GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show (hbox);

  label = gtk_label_new ("Rate:");
  gtk_box_pack_end (GTK_BOX(hbox), label, FALSE, FALSE, 0);
  gtk_widget_show (label);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_table_attach (GTK_TABLE(table), hbox, 1, 2, 1, 2,
		    GTK_FILL|GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show (hbox);

  if (sample != NULL) {
    label = gtk_label_new (g_strdup_printf ("%d kHz",
					    sample->sounddata->format->rate));
    gtk_box_pack_start (GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show (label);
  }

  /* Channels */

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_table_attach (GTK_TABLE(table), hbox, 0, 1, 2, 3,
		    GTK_FILL|GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show (hbox);

  label = gtk_label_new ("Channels:");
  gtk_box_pack_end (GTK_BOX(hbox), label, FALSE, FALSE, 0);
  gtk_widget_show (label);

  hbox = gtk_hbox_new (TRUE, 0);
  gtk_table_attach (GTK_TABLE(table), hbox, 1, 2, 2, 3,
		    GTK_FILL|GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show (hbox);

  if (sample == NULL || sample->sounddata->format->channels == 1) {
    radio = gtk_radio_button_new_with_label (NULL, _("Mono"));
  } else {
    radio = gtk_radio_button_new_with_label (NULL, _("Mono (mixdown)"));
  }
  gtk_box_pack_start (GTK_BOX(hbox), radio, TRUE, TRUE, 0);
  gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)1);
  gtk_signal_connect (GTK_OBJECT(radio), "toggled",
		      set_channels_cb, so);
  gtk_widget_show (radio);

  so->radio_mono = radio;

  if (sample == NULL || sample->sounddata->format->channels == 2) {
    radio = gtk_radio_button_new_with_label
      (gtk_radio_button_group (GTK_RADIO_BUTTON(radio)), _("Stereo"));
  } else {
    radio = gtk_radio_button_new_with_label
      (gtk_radio_button_group (GTK_RADIO_BUTTON(radio)),
       _("Stereo (duplicate)"));
  }
  gtk_box_pack_start (GTK_BOX(hbox), radio, TRUE, TRUE, 0);
  gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)2);
  gtk_signal_connect (GTK_OBJECT(radio), "toggled",
		      set_channels_cb, so);
  gtk_widget_show (radio);

  so->radio_stereo = radio;

  /* Bits */

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_table_attach (GTK_TABLE(table), hbox, 0, 1, 3, 4,
		    GTK_FILL|GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show (hbox);

  label = gtk_label_new ("Bits:");
  gtk_box_pack_end (GTK_BOX(hbox), label, FALSE, FALSE, 0);
  gtk_widget_show (label);

  hbox = gtk_hbox_new (TRUE, 0);
  gtk_table_attach (GTK_TABLE(table), hbox, 1, 2, 3, 4,
		    GTK_FILL|GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show (hbox);

  radio = gtk_radio_button_new_with_label (NULL, _("8 bit"));
  gtk_box_pack_start (GTK_BOX(hbox), radio, FALSE, FALSE, 0);
  gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)8);
  gtk_signal_connect (GTK_OBJECT(radio), "toggled",
		      set_bits_cb, so);
  gtk_widget_show (radio);

  so->radio_8bit = radio;

  radio = gtk_radio_button_new_with_label
    (gtk_radio_button_group (GTK_RADIO_BUTTON(radio)), _("16 bit"));
  gtk_box_pack_start (GTK_BOX(hbox), radio, FALSE, FALSE, 0);
  gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)16);
  gtk_signal_connect (GTK_OBJECT(radio), "toggled",
		      set_bits_cb, so);
  gtk_widget_show (radio);

  so->radio_16bit = radio;

  radio = gtk_radio_button_new_with_label
    (gtk_radio_button_group (GTK_RADIO_BUTTON(radio)), _("24 bit"));
  gtk_box_pack_start (GTK_BOX(hbox), radio, FALSE, FALSE, 0);
  gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)24);
  gtk_signal_connect (GTK_OBJECT(radio), "toggled",
		      set_bits_cb, so);
  gtk_widget_show (radio);

  so->radio_24bit = radio;

  radio = gtk_radio_button_new_with_label
    (gtk_radio_button_group (GTK_RADIO_BUTTON(radio)), _("32 bit"));
  gtk_box_pack_start (GTK_BOX(hbox), radio, FALSE, FALSE, 0);
  gtk_object_set_user_data (GTK_OBJECT(radio), (gpointer)32);
  gtk_signal_connect (GTK_OBJECT(radio), "toggled",
		      set_bits_cb, so);
  gtk_widget_show (radio);

  so->radio_32bit = radio;

  /* OK */

  ok_button = gtk_button_new_with_label (_("OK"));
  GTK_WIDGET_SET_FLAGS (GTK_WIDGET (ok_button), GTK_CAN_DEFAULT);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG(dialog)->action_area), ok_button,
		      TRUE, TRUE, 0);
  gtk_widget_show (ok_button);

  so->ok_button = ok_button;

  /* Cancel */

  button = gtk_button_new_with_label (_("Cancel"));
  GTK_WIDGET_SET_FLAGS (GTK_WIDGET (button), GTK_CAN_DEFAULT);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG(dialog)->action_area), button,
		      TRUE, TRUE, 0);
  gtk_widget_show (button);
  gtk_signal_connect (GTK_OBJECT(button), "clicked",
		      GTK_SIGNAL_FUNC (sndfile_save_options_dialog_cancel_cb),
		      so);

  gtk_widget_grab_default (ok_button);

  update_save_options_caps (so);
  update_save_options_values (so);

  return (dialog);
}

int
sndfile_save_options_dialog (sw_sample * sample)
{
  GtkWidget * dialog;
  sndfile_save_options * so;
  SF_INFO * sfinfo;

  so = g_malloc0 (sizeof(*so));

  so->sample = sample;

  if (sample->file_format == SWEEP_FILE_FORMAT_BY_EXTENSION) {
    sample->file_format = guess_file_format (sample->filename);
  }

  sfinfo = g_malloc (sizeof(SF_INFO));
  memcpy (sfinfo, sample->file_info, sizeof(SF_INFO));

  so->pathname = sample->filename;
  so->sfinfo = sfinfo;

  switch (sample->file_format) {
  case SWEEP_FILE_FORMAT_RAW:
    sfinfo->format = (SF_FORMAT_RAW);
    break;
  case SWEEP_FILE_FORMAT_AIFF:
    sfinfo->format = (SF_FORMAT_AIFF | SF_FORMAT_PCM);
    break;
  case SWEEP_FILE_FORMAT_AU:
    sfinfo->format = (SF_FORMAT_AU | SF_FORMAT_PCM);
    break;
  case SWEEP_FILE_FORMAT_PAF:
    sfinfo->format = (SF_FORMAT_PAF);
    break;
  case SWEEP_FILE_FORMAT_SVX:
    sfinfo->format = (SF_FORMAT_SVX);
    break;
  case SWEEP_FILE_FORMAT_IRCAM:
    sfinfo->format = (SF_FORMAT_IRCAM);
    break;
  case SWEEP_FILE_FORMAT_VOC:
    sfinfo->format = (SF_FORMAT_VOC);
    break;
  case SWEEP_FILE_FORMAT_WAV:
  default:
    sfinfo->format = (SF_FORMAT_WAV | SF_FORMAT_PCM);
    break;
  }

  dialog = create_sndfile_save_options_dialog (so);
  gtk_window_set_title (GTK_WINDOW(dialog), "Save options");

  gtk_signal_connect (GTK_OBJECT(so->ok_button), "clicked",
		      GTK_SIGNAL_FUNC (sndfile_save_options_dialog_ok_cb),
		      so);

  gtk_widget_show (dialog);

  return 0;
}

typedef struct _sf_data sf_data;

struct _sf_data {
  sw_sample * sample;
  SNDFILE * sndfile;
  SF_INFO * sfinfo;
};

static sw_sample *
sample_load_sfinfo (char * pathname, SF_INFO * sfinfo);

static void
sndfile_load_options_dialog_ok_cb (GtkWidget * widget, gpointer data)
{
  sndfile_save_options * so = (sndfile_save_options *)data;
  char * pathname = so->pathname;
  SF_INFO * sfinfo = so->sfinfo;
  GtkWidget * dialog;

  dialog = gtk_widget_get_toplevel (widget);
  gtk_widget_destroy (dialog);

  g_free (so);

  sample_load_sfinfo (pathname, sfinfo);
}

static sw_sample *
sample_load_sf_data (void * data)
{
  sf_data * sf = (sf_data *)data;
  sw_sample * sample = sf->sample;
  SNDFILE * sndfile = sf->sndfile;
  SF_INFO * sfinfo = sf->sfinfo;

  sw_audio_t * d;
  sw_framecount_t remaining, n, run_total;
  gint percent;

  gboolean active = TRUE;

  sf_command  (sndfile, "norm float", "on", 0) ;

  remaining = sfinfo->samples;
  run_total = 0;

  d = sample->sounddata->data;

  while (active && remaining > 0) {
    g_mutex_lock (sample->ops_mutex);

    if (sample->edit_state == SWEEP_EDIT_STATE_CANCEL) {
      active = FALSE;
    } else {
      n = MIN (remaining, 1024);
      sf_readf_float (sndfile, d, n);
      remaining -= n;

      d += (n * sfinfo->channels);

      run_total += n;
      percent = (run_total * 100) / sfinfo->samples;
      sample_set_progress_percent (sample, percent);
    }

    g_mutex_unlock (sample->ops_mutex);
  }

  sf_close (sndfile) ;

  sample_set_edit_state (sample, SWEEP_EDIT_STATE_DONE);

  return sample;
}

static sw_sample *
sample_load_sfinfo (char * pathname, SF_INFO * sfinfo)
{
  SNDFILE *sndfile;
  char buf[BUF_LEN];

  sndfile_save_options * so;
  GtkWidget * dialog;

  sw_sample * sample;
  char * fn, * dn;
  sw_view * v;

  pthread_t load_thread = -1;

  sf_data * sf;

#define RAW_ERR_STR \
"Error while opening RAW file for read. Must specify format, pcmbitwidth and channels."

  if (!(sndfile = sf_open_read (pathname, sfinfo))) {
    sf_error_str (NULL, buf, BUF_LEN);
    if (!strncmp (buf, RAW_ERR_STR, BUF_LEN)) {

      so = g_malloc0 (sizeof(*so));
      so->sample = NULL;
      so->pathname = pathname;
      so->sfinfo = sfinfo;

      sfinfo->format = (SF_FORMAT_RAW | SF_FORMAT_PCM_S8);
      sfinfo->samplerate = 44100;
      sfinfo->pcmbitwidth = 8;
      sfinfo->channels = 2;

      dialog = create_sndfile_save_options_dialog (so);
      gtk_window_set_title (GTK_WINDOW(dialog), "Load Raw PCM options");

      gtk_signal_connect (GTK_OBJECT(so->ok_button), "clicked",
			  GTK_SIGNAL_FUNC (sndfile_load_options_dialog_ok_cb),
			  so);

      gtk_widget_show (dialog);

    } else {
      sf_perror (NULL);
      g_free (sfinfo);
    }

    return NULL;
  }


  /* Split directory and file names
   * RPK: I use pure GLib functions. Its more clear. */
  fn = g_basename(pathname);
  dn = g_dirname(pathname);

  sample = sample_new_empty(dn, fn, sfinfo->channels, sfinfo->samplerate,
			    sfinfo->samples);

  /* Clean up some stuff ... */
  g_free(dn);

  if(!sample) {
    g_free (sfinfo);
    return NULL;
  }

  sample->file_info = sfinfo;

  sample_set_edit_state (sample, SWEEP_EDIT_STATE_BUSY);
  sample_set_edit_mode (sample, SWEEP_EDIT_MODE_FILTER);
  sample_set_progress_percent (sample, 0);

  sample_bank_add(sample);

  v = view_new_all (sample, 1.0);
  sample_add_view (sample, v);

  g_snprintf (buf, BUF_LEN, "Loading %s (%%p%%%% complete)", fn);
  sample_set_progress_text (sample, buf);

  sf = g_malloc0 (sizeof(sf_data));

  sf->sample = sample;
  sf->sndfile = sndfile;
  sf->sfinfo = sfinfo;

  pthread_create (&load_thread, NULL, (void *) (*sample_load_sf_data), sf);

  gtk_timeout_add (30, (GtkFunction)update_edit_progress, (gpointer)sample);

  return NULL;
}

sw_sample *
sndfile_sample_load (char * pathname)
{
  SF_INFO * sfinfo;

  if (pathname == NULL) return NULL;

  sfinfo = g_malloc0 (sizeof (SF_INFO));

  return sample_load_sfinfo (pathname, sfinfo);
}

static int
sndfile_sample_save_thread (void * data)
{
  sf_data * sf = (sf_data *)data;

  sw_sample * sample = sf->sample;
  SNDFILE *sndfile;
  SF_INFO * sfinfo;
  char pathname[SW_DIR_LEN], *p;
  sw_format * format;
  sw_audio_t * fbuf, * d;
  sw_framecount_t nwritten = 0, len, n;
  int i;
  gint percent;

  gboolean active = TRUE;

  if (sample == NULL) return -1;

  if (*sample->directory) {
    snprintf (pathname, SW_DIR_LEN,
	      "%s/%s", sample->directory, sample->filename);
    p = pathname;
  } else {  
    p = sample->filename;
  }

  format = sample->sounddata->format;

  sfinfo = (SF_INFO *)sample->file_info;

  if (sfinfo == NULL) {
    sfinfo = g_malloc0 (sizeof(*sfinfo));
    sample->file_info = sfinfo;
  }

  sfinfo->samplerate  = (int)format->rate;
  sfinfo->samples     = (int)sample->sounddata->nr_frames;

  if (!(sndfile = sf_open_write (p, sfinfo))) {      
    sf_perror (NULL);
    return -1;
  }

  sf_command  (sndfile, "norm float", "on", 0) ;

  if ((int)format->channels == sfinfo->channels) {
    fbuf = sample->sounddata->data;
    while (active && nwritten < sfinfo->samples) {
      g_mutex_lock (sample->ops_mutex);

      if (sample->edit_mode == SWEEP_EDIT_MODE_FILTER) {
	len = MIN (sfinfo->samples - nwritten, 1024);
	n = sf_writef_float (sndfile, fbuf, len);
	fbuf += n * sfinfo->channels;
	nwritten += n;
	percent = (nwritten * 100) / sfinfo->samples;
	sample_set_progress_percent (sample, percent);
      } else {
	active = FALSE;
      }

      g_mutex_unlock (sample->ops_mutex);
    }
  } else if (format->channels == 1 && sfinfo->channels == 2) {
    /* Duplicate mono to stereo */
    fbuf = (sw_audio_t *)alloca(1024 * sizeof(sw_audio_t));
    d = sample->sounddata->data;
    while (active && nwritten < sfinfo->samples) {
      g_mutex_lock (sample->ops_mutex);

      if (sample->edit_mode == SWEEP_EDIT_MODE_FILTER) {
	len = MIN (sfinfo->samples - nwritten, 512);
	for (i = 0; i < len; i++) {
	  fbuf[i*2] = fbuf[i*2+1] = *d++;
	}
	nwritten += sf_writef_float (sndfile, fbuf, len);
	percent = (nwritten * 100) / sfinfo->samples;
	sample_set_progress_percent (sample, percent);
      } else {
	active = FALSE;
      }

      g_mutex_unlock (sample->ops_mutex);
    }
  } else if (format->channels == 2 && sfinfo->channels == 1) {
    /* Mix down stereo to mono */
    fbuf = (sw_audio_t *)alloca(1024 * sizeof(sw_audio_t));
    d = sample->sounddata->data;
    while (active && nwritten < sfinfo->samples) {
      g_mutex_lock (sample->ops_mutex);

      if (sample->edit_mode == SWEEP_EDIT_MODE_FILTER) {
	len = MIN (sfinfo->samples - nwritten, 1024);
	for (i = 0; i < len; i++) {
	  fbuf[i] = *d++;
	  fbuf[i] += *d++;
	  fbuf[i] /= 2.0;
	}
	nwritten += sf_writef_float (sndfile, fbuf, len);
	percent = (nwritten * 100) / sfinfo->samples;
	sample_set_progress_percent (sample, percent);
      } else {
	active = FALSE;
      }

      g_mutex_unlock (sample->ops_mutex);
    }
  } else {
    /* XXX */
  }

  sf_close (sndfile) ;

  return 0;
}

int
sndfile_sample_save (sw_sample * sample)
{
  sf_data * sf;
  pthread_t save_thread = -1;
  char buf[BUF_LEN];

  sample_set_edit_state (sample, SWEEP_EDIT_STATE_BUSY);
  sample_set_edit_mode (sample, SWEEP_EDIT_MODE_FILTER);
  sample_set_progress_percent (sample, 0);

  g_snprintf (buf, BUF_LEN, "Saving %s (%%p%%%% complete)", sample->filename);
  sample_set_progress_text (sample, buf);

  sf = g_malloc0 (sizeof (sf_data));
  sf->sample = sample;

  pthread_create (&save_thread, NULL, (void *) (*sndfile_sample_save_thread),
		  sf);

  gtk_timeout_add (30, (GtkFunction)update_edit_progress, (gpointer)sample);

  return 0;
}

#endif
